'' System Version 2.001
'' OS_fileSystem Changes:
''
''   2006/09/24 - Initial version with change dates
''   2006/10/02 - Improved probeEEPROMs checking
''   2006/10/06 - Improved lores VGA formatting
''   2006/10/28 - Modified use of "BOOT" for Hydra
''   2006/11/02 - Added "noOS" option for files
''   2006/11/03 - Smallest EEPROM device is 64K x 8
''                Only write changed bytes to EEPROM
''                "searchFile" returns options for BOOT
''   2006/11/04 - Added screen full pause to fileList
''   2006/11/06 - Made checkSums PUB for OS_initialization
''   2006/11/09 - Fixed timeout error on initial boot
''   2006/11/12 - probeEEPROMs checks for 32K boot EEPROM

'' This object provides declarations and routines to manipulate
'' named object code spaces in I2C EEPROMS.  It also handles a
'' persistent storage area kept in the last long of the boot
'' EEPROM area.  This is overwritten when a new program is
'' downloaded to the EEPROM by the IDE, but is relocated to the
'' next "page" of the boot EEPROM when this is available.  This
'' area keeps a map of which I/O pins are used for I2C busses.
'' If present, the system automatically probes those busses for
'' EEPROMs during initialization.  There is a copy of the bitmap
'' stored in every code space when it is either named or deleted,
'' but the only copy referenced is the one in the second or first
'' code space of the boot EEPROM.

'' Named object code spaces are currently always 32Kx8 with the
'' name of the program stored in the last 4 longs (16 bytes) of
'' the 32Kx8 "page".  The last long contains system information
'' including a checksum so the file name is actually 12 chars.

'' The bitmap of pins with I2C busses is stored in every initialized
'' "page" for checking purposes although only the one in the boot
'' EEPROM or the next 32K page on the boot EEPROM buss is referenced.
'' A uninitialized "page" is one where at least one checksum doesn't
'' match.  A deleted "page" is one where the checksums are valid and
'' the filename is blank.

'' These routines should be able to support another storage scheme,
'' but others may require some form of space compaction which can
'' be slow with a limited speed serial bus.

'' The Hydra is equipped with a 128K boot EEPROM and can use plug-in
'' 128K EEPROMs that relocate the on-board boot EEPROM to the next
'' 128K address space while taking the place of the boot EEPROM.
'' As a result, the on-board boot EEPROM filename area, while not
'' needed normally, is relocated to a higher address where the
'' boot EEPROM filename area will appear.  The OS has no way of
'' knowing whether the boot EEPROM filename area is on-board or
'' on an expansion card.  To accomodate this, the boot EEPROM
'' filename area is always filled with "**BOOT" and the command
'' interpreter's "copy" command will not create this filename.

CON
  busBitmap   = $8000 - 2         ' Bitmap of pins with I2C busses
  checkMap    = $8000 - 3         ' Checksum for busBitmap
  checkSum    = $8000 - 4         ' Checksum for filename
  filename    = $8000 - 16        ' 12 character filename

  bootName    = $544F4F42         ' BOOT
  blankName   = $20202020         ' blanks
  altBoot1    = $4F422A2A         ' Alternative boot name "**BO", "OT  "
  altBoot2    = $2020544F         ' (Used with Hydra for relocated BOOT)

  opt_text    = %0000_0000_0001   ' file is text (not binary)
  opt_noOS    = %0000_0000_0010   ' disable OS while loading program
                                  ' (program does all its own I/O)

OBJ
  ldr   : "OS_loader"             ' System definitions & I2C driver
  dsp   : "OS_dspText"            ' Display text driver modified for OS use
  key   : "OS_keyboard"           ' Keyboard driver modified for OS use

PUB updateBitmapEEPROMs(addI2C) | i, t, m, x
'' This routine scans all the I2C busses identified in the stored
'' bitmap and fills in the EEPROM table based on what's present.
'' Any I2C busses that do not have EEPROMs are removed from the bitmap.
'' The parameter is either -1 or the number of a pin pair to try adding.
  result := getBusBitmap
  if addI2C <> -1
    result |= |<(addI2C >> 1)     ' Add specified bus
  repeat i from 0 to ldr#maxPinPairs-1
    m := |<i
    if (result & m) <> 0          ' If known, check if present now
      t := probeEEPROMs(i<<1)
      if t == 0
        result &= !m              ' If not present, remove from bitmap
        if (i<<1) == ldr#i2cBootSCL
          abort result            ' Boot EEPROM not present
      word[long[ldr#systemPtr]+ldr#sys_eeproms][i] := t
  updateBitmap(result)

PRI getBusBitmap | i, j, mem, chkM, c, bitmap
'' This routine checks the first 32K of the boot EEPROM for a valid
'' I2C pin bitmap.  If there's more than 32K, the second 32K "page"
'' is checked first.  If there's a valid bitmap, this is returned.
'' Otherwise, a default bitmap showing only the boot EEPROM is used.
  mem := word[long[ldr#systemPtr]+ldr#sys_eeproms+ldr#i2cBootSCL]
  repeat i from 1 to 0
    if (i == 0) or ((mem & |<i) <> 0)
      bitmap~
      chkM := 0
      if not ldr.readEEPROM((ldr#i2cBootSCL<<18)|(i<<15)|checkMap,@chkM,1)
        abort -string("getBusBitmap chkM read err")
      repeat j from 0 to 1
        c := 0
        if not ldr.readEEPROM((ldr#i2cBootSCL<<18)|(i<<15)|(busBitmap+j),@c,1)
          abort -string("getBusBitmap bitM read err")
        bitmap |= c << (8 * j)
        chkM += c
      if (chkM & $FF) == $FF
        return bitmap
  return |<(ldr#i2cBootSCL >> 1)

PRI updateBitmap(bitmap) | i, j, mem, chkM, c, t
'' This routine updates the I2C bus bitmap stored in the boot
'' EEPROM.  If there's a second "page", the bitmap is recorded
'' there.  Otherwise, it's saved in the first (or only) 32K.
  t := 0
  mem := word[long[ldr#systemPtr]+ldr#sys_eeproms+ldr#i2cBootSCL]
  repeat i from 1 to 0
    if (|<i & mem) or (i == 0)
      chkM := $FF
      repeat j from 0 to 1
        c := bitmap & $FF
        bitmap := bitmap >> 8
        chkM -= c
        if not ldr.readEEPROM(ldr#bootAddr|(i<<15)|(busBitmap+j),@t,1)
          abort -string("updateBitmap bitM read err")
        if t <> c                 ' Only write bytes that have changed
          if not ldr.writeEEPROM(ldr#bootAddr|(i<<15)|(busBitmap+j),@c,1)
            abort -string("updateBitmap bitM write err")
          if not ldr.writeWait(ldr#bootAddr)
            abort -string("updateBitmap bitM timeout")
      if not ldr.readEEPROM(ldr#bootAddr|(i<<15)|checkMap,@t,1)
        abort -string("updateBitmap chkM read err")
      if t <> chkM                ' Only write checksum if changed
        if not ldr.writeEEPROM(ldr#bootAddr|(i<<15)|checkMap,@chkM,1)
          abort -string("updateBitmap chkM write err")
        if not ldr.writeWait(ldr#bootAddr)
          abort -string("updateBitmap chkM timeout")
      return

PUB markNewName(namePtr,addr,opt)
'' This routine initializes the code space at the specified address
'' on the I2C bus with the supplied file name.  It does check that
'' the code space is either marked as deleted or is uninitialized.
'' The sign bits of the name are used for 12 option bits given here.
'' The address format is pin pair in bits 22..19, page number in
'' bits 18..15, and address in 14..0.
  addr &= $7FF000                 ' If option bits present, mask them
  if checkSums(0,addr) <> true
    writeFilename(namePtr,addr,opt)
    return false
  abort -string("file not created, space not valid")

PUB markDeleted(namePtr,addr)
'' This routine marks as deleted the code space at the specified
'' address on the I2C bus.  It does check that the code space
'' has the specified name unless the pointer is zero.
'' The sign bits of the name are used for 12 option bits, zeroed here.
'' The address format is pin pair in bits 22..19, page number in
'' bits 18..15, and address in 14..0.
  addr &= $7FF000                 ' If option bits present, mask them
  if checkSums(namePtr,addr) <> true
    writeFilename(0,addr,0)
    return false
  abort -string("file not deleted, space not valid")

PUB searchFile(namePtr) | i, j, b, addr, t, k
'' This routine searches all available EEPROMs for an object code
'' space with the name given.  It returns the pin pair in bits 22-19,
'' the 32K page number in bits 18-15, and the address in bits 14-0.
'' If the code space is not found, an abort is done with the
'' negative address of an error message returned.  If namePtr is
'' zero, this looks for an uninitialized or deleted code space.
'' The sign bits of the name are used for 12 option bits.  These
'' are returned in the least significant 12 bits of the address.
'' The name "BOOT" is a special case for the boot EEPROM.
  if namePtr <> 0
    if (long[namePtr][0] == bootName) and (long[namePtr][1] == blankName)
      return ldr#bootAddr | checkSums(namePtr,ldr#bootAddr)
  repeat i from 0 to ldr#maxPinPairs-1   ' Pin pairs
    b := word[long[ldr#systemPtr]+ldr#sys_eeproms][i]
    repeat j from 0 to 15        ' 32Kx8 spaces
      if (|<j & b) <> 0          ' Is this one present?
        addr := (i << 19) | (j << 15)  ' Check name,chksums        
        if (t := checkSums(namePtr,addr)) <> true
          return addr | t        ' Match, return location
  abort -string("file not found") ' and option bits

PUB checkSums(namePtr, addr) | i, p, c, eq, chkS, chkM
'' This routine checks the filename and bitmap checksums.
'' If those are correct, it compares the filename to the
'' one supplied (or to all blanks if none supplied) and
'' returns the options bits if all match.  If the checksums
'' fail and no name was supplied, a false is returned (to
'' catch the uninitialized situation).  Otherwise return true.
  if (p := namePtr) == 0
    p := string("            ")   ' Use blanks for no file name
  addr &= $7FF000                 ' If option bits present, mask them
  if addr == ldr#bootAddr         ' The boot EEPROM is always "BOOT"
    if long[p][0] == bootName and long[p][1] == blankName
      p := string("**BOOT      ") '  although it's labelled "**BOOT"
    else
      return true                 ' Only BOOT will match
  if addr == ldr#altBootAddr      ' Relocated on-board boot EEPROM
    if long[p][0] <> altBoot1 or long[p][1] <> altBoot2
      return true                 ' Only **BOOT will match
  chkM := 0
  if not ldr.readEEPROM(addr+checkMap,@chkM,1)
    abort -string("checkSums chkM read err")
  repeat i from 0 to 1
    c := 0
    if not ldr.readEEPROM(addr+busBitmap+i,@c,1)
      abort -string("checkSums bitM read err")
    chkM += c
  eq := true
  chkS := 0
  if not ldr.readEEPROM(addr+checkSum,@chkS,1)
    abort -string("checkSums chkS read err")
  repeat i from 0 to 11
    c := 0
    if not ldr.readEEPROM(addr+filename+i,@c,1)
      abort -string("checkSums name read err")
    chkS += c
    eq := eq and ((c & $7F) == byte[p++])
    result |= (c >> 7) << i
  if ((chkS & $FF) == $FF) and ((chkM & $FF) == $FF)
    if not eq
      result := true
  else
    result := namePtr <> 0

PUB writeFilename(namePtr, addr, opt) | i, p, c, chkS, chkM, m, t
'' This routine writes the filename and bitmap fields and their checksums
'' to the end of the 32Kx8 space given by "addr".  The option information
'' in the sign bits of the characters is provided in "opt".
'' The boot EEPROM is always given the name "**BOOT" although it is
'' always referenced as "BOOT".  On the Hydra, where the on-board boot
'' EEPROM can be relocated to the 2nd 128K address space, the on-board
'' boot EEPROM can be referred to as "**BOOT" when it is relocated.
  t := 0
  m := getBusBitmap
  addr &= $7FF000                 ' If option bits present, mask them
  chkM := $FF
  repeat i from 0 to 1            ' Write bitmap bytes & compute checksum
    c := m & $FF
    m := m >> 8
    chkM -= c
    if not ldr.readEEPROM(addr+busBitmap+i,@t,1)
      abort -string("writeFile bitM read err")
    if t <> c                     ' Only write bytes that have changed
      if not ldr.writeEEPROM(addr+busBitmap+i,@c,1)
        abort -string("writeFile bitM write err")
      if not ldr.writeWait(addr)
        abort -string("writeFile bitM timeout")
  if not ldr.readEEPROM(addr+checkMap,@t,1)
    abort -string("writeFile chkM read err")
  if t <> chkM                    ' Only write checksum if it has changed
    if not ldr.writeEEPROM(addr+checkMap,@chkM,1)
      abort -string("writeFile chkM write err")
    if not ldr.writeWait(addr)
      abort -string("writeFile chkM timeout")
  if (p := namePtr) == 0
    p := string("            ")   ' If no name, use blanks
  if addr == ldr#bootAddr or addr == ldr#altBootAddr
    p := string("**BOOT      ")   ' Boot EEPROM actual name (ignore namePtr)
  chkS := $FF
  repeat i from 0 to 11
    c := (byte[p++] & $7F) | ((opt << 7) & $80)
    opt >>= 1                     ' Insert option bits and compute checksum
    chkS -= c
    if not ldr.readEEPROM(addr+filename+i,@t,1)
      abort -string("writeFile name read err")
    if t <> c                     ' Only write bytes that have changed
      if not ldr.writeEEPROM(addr+filename+i,@c,1)
        abort -string("writeFile name write err")
      if not ldr.writeWait(addr)
        abort -string("writeFile name timeout")
  if not ldr.readEEPROM(addr+checkSum,@t,1)
    abort -string("writeFile chkS read err")
  if t <> chkS                    ' Only write checksum if it has changed
    if not ldr.writeEEPROM(addr+checkSum,@chkS,1)
      abort -string("writeFile chkS write err")
    if not ldr.writeWait(addr)
      abort -string("writeFile chkS timeout")

PUB probeEEPROMs(pin) : bitmap | device, b0, b1, b2, b3, c0, c1, c2, c3
'' This routine builds a bitmap showing which 32K x 8 address spaces
'' have been provided if any and returns a value suitable for placing
'' in the system's EEPROM table.  Note that pin must be even.
  repeat device from 14 to 0 step 2              ' Go through the EEPROM devices
    if ldr.checkPresence((pin<<18)|(device<<15)) '  checking each 64K address space
      if pin == ldr#i2cBootSCL and device == 0   ' See if first 64K address space of
        if not ldr.readEEPROM((pin<<18)|$7FF0,@b0,16) ' the boot I2C bus has the same
          abort -string("probeEEPROMs boot read err") '  data in both 32K slots
        if not ldr.readEEPROM((pin<<18)|$FFF0,@c0,16) ' The only way this should happen
          abort -string("probeEEPROMs next read err") '  is when the boot EEPROM is 32K
        if b0 == c0 and b1 == c1 and b2 == c2 and b3 == c3
          bitmap |= |<device                     ' Show just the first 32K as present
          next        
      bitmap |= |<device | |<(device + 1)        ' Mark two 32K banks as present

PUB dumpMemory(pin,addr,size) | i, c, p, page, first, wide, width
'' This routine dumps a portion of the RAM (pin > highest pin) or EEPROM to the
'' display.  The format is 16 bytes wide with hexadecimal and printable characters
'' when the display is 80 columns or wider, otherwise it's 8 bytes wide with
'' hexadecimal only.  If no EEPROM is present at that location, an abort is done
'' supplying the address of an error message.
  if wide := (dsp#cols => 80)
    width := 16
  else
    width := 8
  page := (addr >> 15) & $1F
  addr &= $7FFF
  if pin < (ldr#maxPinPairs << 1)
    if not ldr.readEEPROM((pin<<18)|(page<<15)|addr,@c,1)
      abort string("no EEPROM at that address")
  if addr + size > $7FFF               ' truncate to end of code space
    size := $8000 - addr
  first := true
  p := addr & $7FF0
  repeat while p < (addr + size)
    dsp.str(string("    "))
    if pin < (ldr#maxPinPairs << 1)
      dsp.decFixed(pin,2)
      dsp.out(":")
    else
      dsp.str(string("   "))
    if first
      dsp.hex((page << 15) | addr,5)
      first := false
    else
      dsp.hex((page << 15) | p,5)
    if wide
      dsp.out3(dsp#MoveX,62,"|")
    repeat i from 0 to (width - 1)
      if p => addr and p < (addr + size)
        if pin < (ldr#maxPinPairs << 1)
          if not ldr.readEEPROM((pin<<18)|(page<<15)|p,@c,1)
            abort -string("dumpMemory read err")
        else
          c := byte[p]
        dsp.out2(dsp#MoveX,3*i+13)
        dsp.hex(c,2)
        if wide
          dsp.out2(dsp#MoveX,i+63)
          if c < " " or c > "~"
            dsp.out(".")
          else
            dsp.out(c)
      p++
    if wide
      dsp.out3(dsp#MoveX,79,"|")
    dsp.out(dsp#Cr)
  
PUB fileList | i, j, k, bitmap, chkM, chkS, t, opt, wide, lines, a, b, c
'' This routine displays a list of all files currently known to the system.
  wide := (dsp#cols => 80)
  lines := 0
  repeat i from 0 to 15          ' Pin pairs
    bitmap := word[long[ldr#systemPtr]+ldr#sys_eeproms][i]
    repeat j from 0 to 15        ' 32Kx8 spaces
      if (|<j & bitmap) <> 0     ' Is this one present?
        chkM := 0
        if not ldr.readEEPROM((i<<19)|(j<<15)|checkMap,@chkM,1)
          abort -string("fileList chkM read err")
        repeat k from 0 to 1     ' Make sure bitmap checksum correct
          t := 0
          if not ldr.readEEPROM((i<<19)|(j<<15)|busBitmap+k,@t,1)
            abort -string("fileList bitM read err")
          chkM += t
        chkS := 0
        if not ldr.readEEPROM((i<<19)|(j<<15)|checkSum,@chkS,1)
          abort -string("fileList chkS read err")
        opt := 0                 ' Read filename, extract option bits,
        repeat k from 0 to 11    '  and make sure checksum is correct
          t := 0
          if not ldr.readEEPROM((i<<19)|(j<<15)|filename+k,@t,1)
            abort -string("fileList name read err")
          chkS += t
          byte[@a][k] := t & $7F
          opt |= t >> 7 << k
        if dsp#cols > 36
          dsp.str(string("    "))
        dsp.decFixed(i<<1,2)
        dsp.out(":")
        dsp.hex(j<<15,5)
        if (chkS & $FF) <> $FF or (chkM & $FF) <> $FF
          dsp.str(string(" Unused (uninitialized)",dsp#Cr))
          next
        if a == blankName and b == blankName and c == blankName
          dsp.str(string(" Unused (deleted)",dsp#Cr))
          next
        case (i << 19) | (j << 15)
          ldr#bootAddr:
            dsp.str(string(" Boot EEPROM    [$"))
          ldr#altBootAddr:
            dsp.str(string(" **Boot EEPROM  [$"))
          other:
            dsp.str(string(" <"))
            repeat k from 0 to 11
              dsp.out(byte[@a][k])
            dsp.str(string("> [$"))
        dsp.hex(opt,3)
        if wide and (opt & opt_text)
          dsp.str(string(",text"))
        if wide and (opt & opt_noOS)
          dsp.str(string(",noOS"))
        dsp.str(string("]",dsp#Cr))
        if (lines += 1) == (dsp#rows - 1)
           dsp.str(string("Press any key to continue"))
           key.getkey
           dsp.out(dsp#Cr)
           lines := 0
