'' System Version 2.001
'' OS_mainProgram Changes:
''
''   2006/09/24 - Initialization routine moved to its own object.  This object
''                is now the command interpreter and some support routines and
''                is the main (top) object in the Operating System.  Test commands
''                added for the uALFAT/USBwiz for MMC/SD memory card access.
''   2006/10/02 - Updated some comments
''   2006/10/03 - Improved screen formatting on small screens
''   2006/10/28 - Modified for use with the Hydra system
''   2006/11/03 - Added support for **BOOT** name for boot code space
''   2006/11/04 - Split interpretation of command from main loop
''                Added "option" command.  Finished "run" changes
''   2006/11/05 - Removed "info" command.
''   2006/11/06 - Added "verify" command.
''   2006/11/10 - Added "color" command.  Removed "load" command.
''   2006/11/12 - Added VGA support to "color" command.

'' This is the command interpreter for the Propeller/Hydra "Operating System".
'' This set of objects is intended to provide a basic set of I/O including TV/VGA
'' text , a PS/2 keyboard, and management of one or more I2C memory busses, each
'' containing up to 256K of EEPROM.  One is always the boot I2C bus on I/O pins
'' 28 and 29.  Others can be added on any pair of I/O pins.  Included is a
'' SPIN loader that can load a normal SPIN program from one of the EEPROMs
'' known to the system, from any location in the EEPROM.  The size of the program
'' is stored with the program by the IDE, but this can be overridden.  The I/O
'' drivers are intended to run continuously in 3 or 4 COGs (1 or 2 for display,
'' 1 for the keyboard, and 1 for I2C).  Their work areas (including buffers) are
'' kept at the end of RAM memory so they should not be overwritten when a new
'' program is loaded.  The new program, if it uses these display text, keyboard,
'' and loader objects, should be able to seamlessly use the display and keyboard
'' and transfer control further to other programs and back to this main routine
'' for further interaction.  The loader will stop any COGs not known to it prior
'' to loading a new progam.  Programs can register COGs intended to "persist"
'' between programs.  These must be running assembly language and use a work area
'' that won't be overlaid.  There are methods provided by the loader to do simple
'' work area allocation.

'' There are several objects that are display dependent, OS_initialization,
'' OS_loader, OS_dspText, and OS_dspTextInit.  OS_initialization contains the
'' display I/O pin definitions and also contains the definitions for the PS/2
'' keyboard pins.  OS_loader contains display dependent memory allocation
'' information.  OS_dspText and OS_dspTextInit contain display specific code.
'' There are 3 sets of objects: one for a TV text display (40 x 13), one for
'' a low resolution VGA text display (32 x 15), and one for a high resolution
'' VGA text display.  The high resolution VGA text display can be compiled for
'' 3 different text resolutions: 80 x 40 (640 x 480), 100 x 50 (800 x 600), and
'' 128 x 64 (1024 x 768).  These are defined in OS_dspText as blocks of constants
'' two of which are commented out with the default being 80 x 40.  Simply comment
'' out the active one and uncomment the one for the desired resolution then
'' recompile.  This cannot be used with the 512 x 384 VGA graphics/text driver.
'' It uses too much memory for its bitmap buffer.  This also cannot currently
'' be used with any of the graphics (TV or low resolution VGA) drivers.  A program
'' that uses these drivers can be run by this OS, but the OS drivers have to be
'' shut down (stopped) and the new program has to supply all of its own I/O, then
'' reload the OS when it's done.  The new program can use OS_loaderInit to do so.

'' All routines (except drivers where noted) are (c) 2006 by Michael Green.
'' Some I/O drivers are (c) by Parallax, some with modifications by others.
'' See the notes at the beginning of each source file for specifics.
'' Permission is given to include this or portions of this program in any program
'' for the Parallax Propeller processor as long as proper credit is given.

'' Currently recognized commands:

'' copy <fileFrom> <fileTo>
''     Copy 32K (0-$7FEF) from one file to another.  Create a file if needed.
''     Any options associated with <fileFrom> are copied to <fileTo>.

'' color <palette> {{fore|fgd2|fgd3|back} <number>}}
''     If only <palette> is specified, set the default color palette number.
''     If either a foreground or background color (or both) is given, set that
''     part of the color information for the specified color palette.  If the
''     display driver supports it, do the same for foreground2 and foreground3.

'' delete <file>
''     Delete the specified file

'' dump {<pin>} <addr> <len>
''     Display the specified EEPROM or HUB RAM locations

'' edit
''     Start a simple screen editor.  Quit using the App Key (to the left of Ins)

'' help
''     Display a list of accepted commands

'' list
''     List files available including their location and option bits

'' option <file> {{+|-}<option>} ...
''     Set or clear one or more options associated with the specified file.

'' probe {<pin>}
''     Optionally probe the possible I2C buss on the specified pin pair for EEPROMs
''     List all known I2C busses and the total amount of EEPROM space on each

'' rename <fileFrom> <fileTo>
''     Rename <fileFrom> to be <fileTo>.  Options are retained.

'' run <file>
''     Execute the SPIN program contained in the specified file

'' verify <file>
''     Perform a checksum on the specified file and report if non-zero.

CON
  _clkmode    = xtal1 + pll8x                  ' Hydra comes supplied with 10MHz crystal
  _xinfreq    = 10_000_000 + 0_000             ' This can be adjusted for actual frequency
  
OBJ
  init    : "OS_initialization"                ' OS Initialization routines and definitions
  ldr     : "OS_loader"                        ' SPIN program loader and support routines
  dsp     : "OS_dspText"                       ' Display text driver modified for OS use
  key     : "OS_keyboard"                      ' PS/2 keyboard driver modified for OS use
  files   : "OS_fileSystem"                    ' File system support
  edit    : "OS_windowEditor"                  ' Screen and file editor

CON
  _free = (ldr#endMemory - ldr#endFree + ldr#totalNeeded + 3) / 4  ' # longs needed

  #1
  processProbe,  processColor,  processRun     ' For command post-processing
  processHelp,   processEdit,   processList
  processCopy,   processDump,   processDelete
  processOption, processRename, processVerify

VAR
  long dumpPin, dumpAddr, dumpSize
  long fileAddr, fileAddrTo, fileChange
  long fileName[3], fileNameTo[3]

DAT
title  byte  "Hydra OS - Version 2.001",dsp#Cr
       byte  "Type help for a command list",dsp#Cr
       byte  0

PUB main | t, y
'' Initialize the operating system, then repeatedly request a one line
'' command from the user (in a single line window), then execute it in
'' the full screen window.  If the value returned by "enterDoCommand"
'' is negative, its absolute value is the address of an error message.
'' "processCommand" returns a non-zero value if there's an error and
'' the absolute value is used as the address of an error message string.
  init.start(@title)
  repeat 
    y := dsp.windowRow
    t := \enterDoCommand(y)
    dsp.out2(dsp#UseWin,0)
    dsp.out4(dsp#MoveXY,dsp#cols,y,dsp#Cr)
    if t > 0
      t := \processCommand(t)
    if t <> 0
      dsp.str(string("*** "))
      dsp.str(||t)
      dsp.out(dsp#Cr)

PRI processCommand(t) | i, j, k, y, z, b0, b1, b2, b3, b4, b5, b6, b7
'' Process the individual commands.  Return zero if successful.  Return
'' or abort with a non-zero value if an error occurs (the address of an
'' error message string).
  case t
    processProbe:
      if dumpPin < (ldr#maxPinPairs << 1)
        if (\files.updateBitmapEEPROMs(dumpPin) & $4000) == 0
          abort string("no boot EEPROM found")
      repeat i from 0 to (ldr#maxPinPairs<<1)-2 step 2
        z~
        y := word[long[ldr#systemPtr]+ldr#sys_eeproms+i]
        if y <> 0
          repeat j from 0 to 15
            if (|<j & y) <> 0
              z += 32
          dsp.str(string("    Pin "))
          dsp.decFixed(i,2)
          dsp.str(string("  "))
          dsp.decFixed(z,3)
          dsp.str(string("K",dsp#Cr))
    processEdit:
      editWindow
    processHelp:
      dsp.str(string(" color <#> {fore|back}<value>",dsp#Cr))
      dsp.str(string(" copy <fileFrom> <fileTo>",dsp#Cr))
      dsp.str(string(" delete <file>",dsp#Cr))
      dsp.str(string(" dump {<pin>} <addr> <len>",dsp#Cr))
      dsp.str(string(" edit,  help,  list",dsp#Cr))
      dsp.str(string(" option <file> {{+|-}<option>}",dsp#Cr))
      dsp.str(string(" probe {<pin>}",dsp#Cr))
      dsp.str(string(" rename <fileFrom> <fileTo>",dsp#Cr))
      dsp.str(string(" run <file>",dsp#Cr))
      dsp.str(string(" verify <file>",dsp#Cr))
    processList:
      files.fileList
    processDump:
      files.dumpMemory(dumpPin,dumpAddr,dumpSize)
    processCopy:
      if fileChange
        dsp.str(string("    Overwrite? (Y/n) "))
        repeat
          t := key.getkey
          if t == dsp#Cr
            t := "y"
        until t == "Y" or t == "y" or t == "N" or t == "n"
        dsp.out(t)
      else
        files.markNewName(@fileNameTo,fileAddrTo,fileAddr)
        t := "y"
        dsp.str(string("Created"))
      if t == "Y" or t == "y"
        if fileChange
          files.writeFilename(@fileNameTo,fileAddrTo,fileAddr)
        dsp.out(" ")
        j := dsp.windowCol
        fileAddr &= $7FF000    ' Remove option bits
        fileAddrTo &= $7FF000  '  from addresses
        repeat i from 0 to $7FE0 step 32
          dsp.decFixed(i,5)
          dsp.out2(dsp#MoveX,j)
          if i == $7FE0
            k := 16
          else
            k := 32
          if not ldr.readEEPROM(fileAddr+i,@b0,k)
            abort string("copy EEPROM read error")
          if not ldr.writeEEPROM(fileAddrTo+i,@b0,k)
            abort string("copy EEPROM write error")
          ldr.writeWait(fileAddrTo+i)
        dsp.str(string("Saved"))
      dsp.out(dsp#Cr)
    processRun:
      if fileAddr & files#opt_text
        abort string("not a program")
      dsp.str(string("    Loading ... ",dsp#Cr))
      repeat i from 0 to 7  ' Stop any cogs not registered
        if (word[long[ldr#systemPtr]+ldr#sys_cog] & |<i) == 0
          cogstop(i)
      ldr.bootEEPROM(fileAddr)             ' This will stop the present cog
      reboot                               ' Reboot if control returns here
    processDelete:
      dsp.str(string("    Are you sure? (Y/n) "))
      repeat
        t := key.getkey
        if t == dsp#Cr
          t := "y"
      until t == "Y" or t == "y" or t == "N" or t == "n"
      dsp.out(t)
      if t == "Y" or t == "y"
        files.markDeleted(@fileName,fileAddr)
        dsp.str(string(" - Done"))
      dsp.out(dsp#Cr)
    processOption:
      t := fileAddr & $FFF        ' Get option bits
      t &= !(fileChange >> 16)    ' Remove those requested
      t |= fileChange & $FFFF     ' Add those requested
      if \files.writeFilename(@fileName,fileAddr,t) < 0
        abort -string("can't write file options")
    processRename:
      if \files.writeFileName(@fileNameTo,fileAddr,fileAddr) < 0
        abort -string("can't rename file")
    processVerify:
      if fileAddr & files#opt_text
        abort string("not a program")
      dsp.str(string("    Verifying ... "))
      if not ldr.verifyEEPROM(fileAddr)
        dsp.str(string("Failed",dsp#Cr))
      else
        dsp.str(string("Done",dsp#Cr))
    processColor:
      if dumpSize < 0
        t := @dumpAddr
      else
        t := @dumpPin
      repeat ||dumpSize
        dsp.out(byte[t++])

PUB editWindow | t
'' Set up a simple screen editor with numbered blank lines.  Use the "App" key
'' to exit the editor (this is left of the "Ins" key on the Parallax keyboard).
  dsp.out6(dsp#SetWin,1,5,1,dsp#cols-5,dsp#rows-2)
  dsp.out6(dsp#SetWin,2,0,1,5,dsp#rows-2)
  dsp.out5(dsp#UseWin,2,dsp#WinColor,0,dsp#ClrScr)
  repeat t from 1 to dsp#rows - 2
    dsp.out3(dsp#MoveXY,0,t-1)
    dsp.decFixed(t,4)
  dsp.out3(dsp#UseWin,1,dsp#ClrScr)
  repeat
    t := edit.windowEditor
  until t == $CC
  dsp.out3(dsp#UseWin,0,dsp#ClrScr)

PUB enterDoCommand(y) | t, x, a, b, c
'' Set up a window at the line specified and accept a command
'' line terminated by a Cr.  Scan for a command word and
'' command specific parameters, either words or numbers.
'' If the command is known and the parameters are valid,
'' execute the command.  Abort with a negative error string
'' address or return a positive command specific value.
  t := dsp.colorPalette
  dsp.out6(dsp#SetWin,1,1,y,dsp#cols-1,1)
  dsp.out6(dsp#UseWin,0,dsp#MoveXY,0,y,">")
  dsp.out5(dsp#UseWin,1,dsp#WinColor,t,dsp#ClrScr)
  repeat
    t := edit.windowEditor
  until t == dsp#Cr
  x := \getTextWord(0,@a,4,true," ")
  if x < 0
    abort string("no command given")
  case a
    $59504F43:  ' copy <fileFrom> <fileTo>
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <fileFrom>")
      x := \getTextWord(x,@fileNameTo,12,true," ")
      if x < 0
        abort -string("missing <fileTo>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<fileFrom> not found")
      fileAddrTo := \files.searchFile(@fileNameTo)
      if fileAddrTo < 0
        if fileNameTo[0] == files#altBoot1
          if fileNameTo[1] == files#altBoot2
           abort -string("can't create **BOOT")
        fileAddrTo := \files.searchFile(0)
        if fileAddrTo < 0
          abort -string("can't create <fileTo>")
        fileChange := false
      else
        fileChange := true
      return processCopy     ' Options were copied
    $424F5250:  ' probe {<pin>}
      x := \getNumber(x,@dumpPin)
      if x < 0
        dumpPin := ldr#maxPinPairs << 1
        return processProbe
      if (dumpPin < (ldr#maxPinPairs << 1)) and (dumpPin & 1 == 0)
        return processProbe
      else
        abort -string("pin number odd or does not exist")
    $204E5552:  ' run <file>
      if (x := \getTextWord(x,@fileName,12,true," ")) < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      return processRun
    $54494445:  ' edit
      return processEdit
    $504C4548:  ' help
      return processHelp
    $5453494C:  ' list
      return processList
    $454C4544:  ' delete <file>
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <file>")
      if fileName[0] == files#bootName
        if fileName[1] == files#blankName
          abort -string("can't delete BOOT")
      if fileName[0] == files#altBoot1
        if fileName[1] == files#altBoot2
          abort -string("can't delete **BOOT")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      return processDelete
    $504D5544:  ' dump {<pin>} <addr> <size>
      x := \getNumber(x,@a)
      if x < 0
        abort -string("missing <pin> or <addr>")
      x := \getNumber(x,@b)
      if x < 0
        abort -string("missing <addr> or <len>")
      x := \getNumber(x,@c)
      if x < 0
        if (a > $7FFFF) or (b => 1024)
          abort -string("invalid number")
        dumpPin := ldr#maxPinPairs << 1
        dumpAddr := a
        dumpSize := b
        return processDump
      else
        if (a => (ldr#maxPinPairs << 1)) or (a & 1) or (b > $7FFFF) or (c => 1024)
          abort -string("invalid number")
        dumpPin := a
        dumpAddr := b
        dumpSize := c
        return processDump
    $4954504F:  ' option <file> {{+|-}<option>} ...
      if (x := \getTextWord(x,@fileName,12,true," ")) < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      fileChange := 0
      repeat until (x := \getTextWord(x,@a,4,true," ")) < 0
        case a
          $5845542B: ' +TEX
            fileChange |= ldr#opt_text
          $5845542D: ' -TEX
            fileChange |= ldr#opt_text << 16
          $4F4F4E2B: ' +NOO
            fileChange |= ldr#opt_noOS
          $4F4F4E2D: ' -NOO
            fileChange |= ldr#opt_noOS << 16
      return processOption
    $414E4552:  ' rename <fileFrom> <fileTo>
      x := \getTextWord(x,@fileName,12,true," ")
      if x < 0
        abort -string("missing <fileFrom>")
      x := \getTextWord(x,@fileNameTo,12,true," ")
      if x < 0
        abort -string("missing <fileTo>")
      if fileName[0] == files#bootName
        if fileName[1] == files#blankName
          abort -string("can't change BOOT")
      if fileName[0] == files#altBoot1
        if fileName[1] == files#altBoot2
          abort -string("can't change **BOOT")
      if fileNameTo[0] == files#bootName
        if fileNameTo[1] == files#blankName
          abort -string("can't change name to BOOT")
      if fileNameTo[0] == files#altBoot1
        if fileNameTo[1] == files#altBoot2
          abort -string("can't change name to **BOOT")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<fileFrom> not found")
      fileAddrTo := \files.searchFile(@fileNameTo)
      if fileAddrTo => 0
        abort -string("<fileTo> already exists")
      return processRename
    $49524556:  ' verify <file>
      if (x := \getTextWord(x,@fileName,12,true," ")) < 0
        abort -string("missing <file>")
      fileAddr := \files.searchFile(@fileName)
      if fileAddr < 0
        abort -string("<file> not found")
      return processVerify
    $4F4C4F43:  ' color <#> {fore|fgd2|fgd3|back}<value>
      x := \getNumber(x,@a)
      if x < 0 or a < 0 or a > 15
        abort -string("missing or invalid <palette>")
      dumpPin.byte[0] := dsp#Color
      dumpPin.byte[1] := a
      dumpPin.byte[2] := dsp.foregroundColor(a)
      dumpPin.byte[3] := dsp.backgroundColor(a)
      dumpAddr.byte[0] := dsp#SetColor2
      dumpAddr.byte[1] := a
      dumpAddr.byte[2] := dsp.foreground3Color(a)
      dumpAddr.byte[3] := dsp.foreground2Color(a)
      t := 0
      repeat until (x := \getTextWord(x,@y,4,true," ")) < 0
        dumpPin.byte[0] := dsp#SetColor
        x := \getNumber(x,@a)
        if x < 0 or (a & !$FF) <> 0
          abort -string("missing or invalid <color>")
        case y
          $45524F46:  ' fore
            dumpPin.byte[2] := a
            t |= %101
          $32444746:  ' fgd2
            dumpAddr.byte[3] := a
            t |= %110
          $33444746:  ' fgd3
            dumpAddr.byte[2] := a
            t |= %110
          $4B434142:  ' back
            dumpPin.byte[3] := a
            t |= %101
          other:
            abort -string("invalid keyword")
      case t
        %000: dumpSize := 2
        %101: dumpSize := 4
        %110: dumpSize := -4
        %111: dumpSize := 8
      return processColor
'    $54534554:  ' test <function>
'      x := \getTextWord(x,@a,8,true," ")
'      if x < 0
'        abort -string("missing <function>")
'      case a
'        $4E414843:  ' change
'        $43454843:  ' check
'        $534F4C43:  ' close
'        $4E4E4F43:  ' connect
'        $59504F43:  ' copy
'        $454C4544:  ' delete
'        $4C4C4946:  ' fill
'        $53554C46:  ' flush
'        $5453494C:  ' list
'        $44414F4C:  ' load
'        $4E45504F:  ' open
'        $44414552:  ' read
'        $45534552:  ' reset
'        $43415453:  ' stack
'        $53524556:  ' version
'        $54495257:  ' write
'        other:
'          abort -string("unknown <function>")
    other:
      abort -string("unknown command")

PRI skipBlanks(x)
'' Skip blanks until at right border of window and return x
'' of first non-blank or abort with a negative error string
'' address.
  repeat
    if dsp.windowAtEnd(x,0)
      abort -string("missing non-blank")
  until dsp.windowChar(x++,0) <> " "
  return x - 1

PRI getTextWord(x,textAddr,maxSize,uppercase,clearFirst) | c
'' Skip leading blanks, then copy up to maxSize non-blank
'' characters into the area at textAddr.  Return the
'' position of the next blank.  If uppercase is true,
'' store any lower case alphabetic characters as upper
'' case.  If clearFirst is not -1 (true), clear the
'' textAddr area to the value given.  If an error occurs,
'' abort with a negative error string address
  if clearFirst <> -1
    repeat c from 1 to maxSize
      byte[textAddr + c - 1] := clearFirst
  x := skipBlanks(x)
  repeat while (c := dsp.windowChar(x++,0)) <> " "
    if maxSize > 0
      if uppercase and c => "a" and c =< "z"
        c := c & !%0100000
      byte[textAddr++] := c
      maxSize--
  return x - 1

PRI getNumber(x,numberAddr) | negative, c, t
'' Skip leading blanks, then convert a decimal number
'' with an optional sign ("+"/"-") into binary to be
'' stored in the long at numberAddr.  Return the
'' position of the next blank.  If an error occurs,
'' abort with a negative error string address.
'' Currently, there's no checking for 32 bit overflow.
  x := skipBlanks(x)
  c := dsp.windowChar(x++,0)
  negative := c == "-"
  if negative or c == "+"
    c := dsp.windowChar(x++,0)
  if c == "$"
    c := dsp.windowChar(x++,0)
    t := hexDigit(c)
    if t < 0
      abort -string("invalid hex character")
    long[numberAddr] := t
    c := dsp.windowChar(x++,0)
    repeat until (t := hexDigit(c)) < 0
      long[numberAddr] := long[numberAddr] << 4 | t
      c := dsp.windowChar(x++,0)
    if negative
      long[numberAddr] := -long[numberAddr]
    return x - 1
  if c == "%"
    c := dsp.windowChar(x++,0)
    if not (c == "0" or c == "1")
      abort -string("invalid hex character")
    long[numberAddr] := c - "0"
    c := dsp.windowChar(x++,0)
    repeat while c == "0" or c == "1"
      long[numberAddr] := long[numberAddr] << 1 | (c - "0")
      c := dsp.windowChar(x++,0)
    if negative
      long[numberAddr] := -long[numberAddr]
    return x - 1
  if not (c => "0" and c=< "9")
    abort -string("invalid numeric character")
  long[numberAddr] := c - "0"
  c := dsp.windowChar(x++,0)
  repeat while c => "0" and c =< "9"
    long[numberAddr] := long[numberAddr] * 10 + c - "0"
    c := dsp.windowChar(x++,0)
  if negative
    long[numberAddr] := -long[numberAddr]
  return x - 1

PRI hexDigit(c)
'' Convert a hexadecimal character to the corresponding value or -1 if invalid.
  if c => "0" and c =< "9"
    return c - "0"
  if c => "A" and c =< "F"
    return c - "A" + 10
  if c => "a" and c =< "f"
    return c - "a" + 10
  return -1
  